#
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
# 
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $
# $Revision: 1.6.2.56 $
# $Source: /project/arrayrep/src/mirrorviewsnap/unix/MirrorViewSnap.pm,v $
#

package MirrorViewSnap;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use warnings;

use POSIX qw(uname);
use Socket;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);

@EXPORT = qw(
        );
$VERSION = '1.00';

require 5.000;
use ag_i18n_inc;

if ( eval "require Net::IP" ) {
    require Net::IP;
}

my (@CMDS, @LINKCMDS, $_DO, $_DOE);

my $SYSTEM  = (uname())[0];
my $nodename  = (uname())[1];

my ($GREP, $AWK);
if ($SYSTEM eq "Linux") {
    $GREP = "/bin/grep";
    $AWK = "/bin/awk";
} else {
    $GREP = "/usr/bin/grep";
    $AWK = "/usr/bin/awk";
}

my $VXASSIST = "/usr/sbin/vxassist";
my $VXDISK = "/usr/sbin/vxdisk";
my $VXDCTL = "/usr/sbin/vxdctl";
my $VXDG = "/usr/sbin/vxdg";
my $VXDDLADM = "/usr/sbin/vxddladm";
my $VXDMPADM = "/opt/VRTS/bin/vxdmpadm";
my $VXCLUSTADM = "/etc/vx/bin/vxclustadm";

#Scan devices
my $DEVFSADM = "/usr/sbin/devfsadm";
my $CFGADM = "/usr/sbin/cfgadm";
my $VXDISKCONFIG = "/usr/sbin/vxdiskconfig";
my $IOSCAN = "/usr/sbin/ioscan";
my $CFGMGR = "/usr/sbin/cfgmgr";
my $UDEV = "/sbin/udev";
my $DEVNULL = "/dev/null";


my $POWERMT = ($SYSTEM eq "AIX") ? "/usr/sbin/powermt" : "/etc/powermt";
my $POWERCF = ($SYSTEM eq "AIX") ? "/usr/sbin/powercf" : "/etc/powercf";
my $LUXADM = "/usr/sbin/luxadm";
my $MODINFO = "/usr/sbin/modinfo";
my $GENKEX = "/usr/bin/genkex";

my ($navisecpath, $naviclipath);

#
# Map of VM technology to name of attribute that contains the
# diskgroup/volumegroup name
#
my %vmattrib = (
	LVMVG => 'VolumeGroup',
	DiskGroup => 'DiskGroup',
	CVMVolDg => 'CVMDiskGroup',
    VMDg => 'DiskGroupName'
);

sub new {
    my($self) = shift;
    my($resname) = shift;
    my($storageGrpName) = shift;
    my($target) = shift;

# get DG name
    my $group = '';
    my @dgname = ();
    my $navisecclihome = '';
    my $numLocalarraySPNames = '';
    my @LocalarraySPNames = ();
    my $mode = '';
    my $cgrpname = '';
    my $numMirrors = '';
    my @mirnames = ();
    my @Luns = ();
    my @SSdisks = ();
    my @PhysicalDevices = ();
    my $lunowner = '';
    my $ownersp = '';
    my $i = 0;
    my @SSOSHandles = '';
    my $localsecfilepath = "";
    my $localsecfileopt = "";
    my $remotesecfileopt = "";

    my $rac = 0;
    my ($tmp) = split('\n', `$HARES -value $target Group`);
    $group = $tmp;
    my ($pll) = split('\n', `$ag_i18n_inc::HAGRP -value $group Parallel`);
    if ($pll == 1) {
        $rac = 1;
    }

    my (%mir_luns) = ();
    my (%lunowner_sp) = ();

    my $sessionName = "VCS" . "_" . "$resname" . "_" . "fd_session";
    my ($typ) = split('\n', `$HARES -value $target Type`);
    my $index = 0;
    my ($vmtype, $dgattrib) = ("", "");
    if ($typ eq "MirrorView") {
        my (@depfields) = split('\n', `$HARES -dep $target`);
        my (@depline) = split('\s+', $depfields[$#depfields]);
        foreach my $depline (@depfields) {
            my (@deppair) = split('\s+', $depline);
            if ($deppair[2] eq $target) {
                ($tmp) = split('\n', `$ag_i18n_inc::HARES -value $deppair[1] Type`);
                if (0 == length $vmtype) {
                    $vmtype = $tmp;
                    if (!exists($vmattrib{$vmtype})) {
                        VCSAG_LOG_MSG("E", "Unsupported type of volume manager $tmp", 50, "$tmp");
                        return undef;
                    }
                    $dgattrib = $vmattrib{$vmtype};
                } else {
                    if ($vmtype ne $tmp) {
                        VCSAG_LOG_MSG("E", "Mixed Volume Management types are not supported; Exiting", 131);
                        return undef;
                    }
                }
                ($tmp) = split('\n', `$HARES -value $deppair[1] $dgattrib`);
                $dgname[$index] = $tmp;
                $index = $index + 1;
            }
        }
        ($tmp) = split('\n', `$HARES -value $target NaviCliHome`);
        if ($^O =~ /^MSWin/i) {
            $navisecclihome = Win32::GetShortPathName($tmp);
            $navisecpath = "$navisecclihome\\naviseccli" if ($^O =~ /^MSWin/i);
            $naviclipath = "$navisecclihome\\navicli" if ($^O =~ /^MSWin/i);

            my $VMPATH = $ENV{"VMPATH"};
            $VMPATH = Win32::GetShortPathName($VMPATH);
            $VXDG = $VMPATH . "vxdg";
            $VXDISK = $VMPATH . "vxdisk";
            $VXASSIST = $VMPATH . "vxassist";
            $VXDG = "\"$VXDG\"";
            $VXDISK = "\"$VXDISK\"";
            $VXASSIST = "\"$VXASSIST\"";

        } else {
            $navisecclihome = $tmp;
            $navisecpath = "$navisecclihome/naviseccli";
            $naviclipath = "$navisecclihome/navicli";
        }
        my @tmp_arr = split('\s+', `$HARES -value $target LocalArraySPNames`);
        $numLocalarraySPNames = ($#tmp_arr + 1);
        @LocalarraySPNames = @tmp_arr;
        ($tmp) = split('\n', `$HARES -value $target Mode`);
        $mode = $tmp;
        ($tmp) = split('\n', `$HARES -value $target GrpName`);
        $cgrpname = $tmp;
        @tmp_arr = split('\s+', `$HARES -value $target MirNames`);
        $numMirrors = ($#tmp_arr + 1);
        if (($cgrpname ne "") && ($numMirrors > 0)) {
            VCSAG_LOG_MSG("W", "You can either specify consistent group name or individual mirrors. Please specify either one for $target.", 409, "$target");
            exit(0);
        }
        @mirnames = @tmp_arr;
        ($tmp) = split('\n', `$HARES -value $target LocalArraySecurityFilePath`);
        $localsecfilepath = $tmp;
        if ($tmp ne "") {
            $localsecfileopt = " -secfilepath \"$tmp\" ";
        }
        ($tmp) = split('\n', `$HARES -value $target RemoteArraySecurityFilePath`);
        $remotesecfileopt = $tmp;
        if ($tmp ne "") {
            $remotesecfileopt = " -secfilepath \"$tmp\" ";
        }

# populate our mirrorNames array
        for($i=0; $i<$numMirrors; $i++){
            $mir_luns{$mirnames[$i]} = "";
        }

        if  ($cgrpname ne "") {
            my $mirror = '';
            my $cmd = "$navisecpath -h $LocalarraySPNames[0] $localsecfileopt mirror -$mode -listgroups -name $cgrpname -mirrors";
            my @query = split('\n', `$cmd`);
            foreach my $q (@query) {
                if ($q =~ "Mirror Name"){
                    my @mirrors = split('\s+', $q);
                    $mirror = $mirrors[2];
                    push(@mirnames, $mirror);
                } elsif ($q =~ "Local LUN Number"){
                    my @luns = split('\s+', $q);
                    $mir_luns{$mirror} = $luns[3];

                    my $ownersp = getlunowner($#LocalarraySPNames, @LocalarraySPNames, $mir_luns{$mirror}, $localsecfileopt);
                    $lunowner_sp{$mir_luns{$mirror}} = $ownersp;
                }
                next;
            }
        } else {
            for($i=0; $i<$numMirrors; $i++){
                my $mirror = $mirnames[$i];
                my $cmd = "$navisecpath -h $LocalarraySPNames[0] $localsecfileopt mirror -$mode -list -name $mirror";
                my @query = split('\n', `$cmd`);
                foreach my $q (@query) {
                    if ($q !~ "Logical Unit Numbers"){
                        next;
                    }

                    my @luns = split('\s+', $q);
                    $mir_luns{$mirror} = $luns[3];

                    my $ownersp = getlunowner($#LocalarraySPNames, @LocalarraySPNames, $mir_luns{$mirror}, $localsecfileopt);
                    $lunowner_sp{$mir_luns{$mirror}} = $ownersp;

                    next;
                }
            }
        }
    } else {
        my (@mvres) = split('\n', `$HARES -list Type=MirrorView Group=$group`);
        my (@mvresn) = split('\s+', $mvres[0]);
        ($tmp) = split('\n', `$HARES -value $mvresn[0] NaviCliHome`);
        if ($^O =~ /^MSWin/i) {
            $navisecclihome = Win32::GetShortPathName($tmp);
            $navisecpath = "$navisecclihome\\naviseccli" if ($^O =~ /^MSWin/i);
            $naviclipath = "$navisecclihome\\navicli" if ($^O =~ /^MSWin/i);
        } else {
            $navisecclihome = $tmp;
            $navisecpath = "$navisecclihome/naviseccli";
            $naviclipath = "$navisecclihome/navicli";
        }
        my @tmp_arr = split('\s+', `$HARES -value $mvresn[0] LocalArraySPNames`);
        $numLocalarraySPNames = ($#tmp_arr + 1);
        @LocalarraySPNames = @tmp_arr;
        ($tmp) = split('\n', `$HARES -value $target DiskGroup`);
        $dgname[0] = $tmp;
        my @tmp = split('\n', `$HARES -value $mvresn[0] LocalArraySecurityFilePath`);
        $localsecfilepath = $tmp[0];
        if ($tmp[0] ne "") {
            $localsecfileopt = " -secfilepath \"$tmp[0]\" ";
        }

        my (%LogicalUnits);
        my (%vxdisk);
        my $lun = '';

# Identify the disk(s) with the source disk group.
# Once a snapshot is taken, the snapshot disk(s) will be the target disks,
# which should be used when importing the disk group.

# Get the vol mgr device names for the disk group
        foreach my $dg (@dgname) {
            my $cmd = "$VXDISK -o alldgs list | $GREP -w udid_mismatch | $GREP -w '$dg'";
            VCSAG_LOG_MSG("I", "Issuing udid_mismatch precheck command: $cmd", 136, "$cmd");
            my (@lines) = split('\n', `$cmd`);
            foreach my $l (@lines) {
                my (@fields) = split('\s+', $l);
                push(@SSdisks, $fields[0]);
            }

            my $arrcount = scalar @SSdisks;
            if ( $arrcount == 0 ) {
                $cmd = "$VXDISK -o alldgs list | $GREP -w '$dg'";
                VCSAG_LOG_MSG("I", "Issuing alldgs command: $cmd", 136, "$cmd");

                my (@lines) = split('\n', `$cmd`);
                foreach my $l (@lines) {
                    my (@fields) = split('\s+', $l);
                    push(@SSdisks, $fields[0]);
                }
            }
        }

# Fetch o/p of vxdisk -p list and create a map from diskname to serial no. This
# serial no is the same as the lun uid given by the array for clariion.
        my $cmd = "$VXDISK -p list";
        my @vxdisk_plist = split('\n', `$cmd`);
        my $disk = "";
        my %disk_serial;
        foreach my $line (@vxdisk_plist) {
            if ($line =~ /^DISK\s+:\s+(\S+)$/) {
                $disk = $1;
                next;
            }
            next if ($disk eq "");
            if ($line =~ /^LUN_SERIAL_NO\s+:\s+(\S+)$/) {
                $disk_serial{$disk}{'serial'} = $1;
                $disk = "";
                next;
            }
        }

# Fetch the o/p of naviseccli getlun -uid and create a map from uid to lun
# number.
        $cmd = "$navisecpath -h $LocalarraySPNames[0] $localsecfileopt getlun -uid";
        my @lun_uid = split('\n', `$cmd`);
        my %uid_lun;
        foreach my $line (@lun_uid) {
            if ($line =~ /^LOGICAL UNIT NUMBER\s+(\d+)$/) {
                $lun = $1;
                next;
            }
            next if ($lun eq "");
            if ($line =~ /UID:\s+(\S+)$/) {
                my $uid = $1;
                $uid =~ s/://g;
                $uid_lun{$uid} = $lun;
                $lun = "";
            }
        }
        foreach my $ssdisk (@SSdisks) {
            next if (!exists($disk_serial{$ssdisk}));
            my $serial = $disk_serial{$ssdisk}{'serial'};
            next if (!exists($uid_lun{$serial}));
            push (@Luns, $uid_lun{$serial});
        }

        for($i=0; $i<$#Luns + 1; $i++){
            my $ownersp = getlunowner($#LocalarraySPNames, @LocalarraySPNames, $Luns[$i], $localsecfileopt);
            $lunowner_sp{$Luns[$i]} = $ownersp;
        }
    }

    my($arrayname)="NOARRAYSP";
    $i=0;
    my $pingretry = 4;
    for($i=0; $i<$numLocalarraySPNames; $i++){
        my $pingret = &PingTest($LocalarraySPNames[$i]);
        if ($pingret) {
            my $j = 0;
            while ( $pingret && ($j < $pingretry )) {
                sleep 1;
                $pingret = &PingTest($LocalarraySPNames[$i]);
                $j++
            }
            if ($pingret  && ($j == $pingretry)) {
                VCSAG_LOG_MSG("N", "Local array SP $LocalarraySPNames[$i] is unreachable.", 1, $LocalarraySPNames[$i] );
            } else {
                $arrayname = $LocalarraySPNames[$i];
                last;
            }
        } else {
            $arrayname = $LocalarraySPNames[$i];
            last;
        }
    }

    $self = {
        "resname" => $resname,
        "group" => $group,
        "storageGrpName" => $storageGrpName,
        "arrayname" => $arrayname,
        "navisecclihome" => $navisecclihome,
        "mode" => $mode,
        "cgrpname" => $cgrpname,
        "numMirrors" => $numMirrors,
        "flag" => $mode,
        "sessionName" => $sessionName,
        "localsecfileopt" => $localsecfileopt,
        "remotesecfileopt" => $remotesecfileopt,
        "rac" => $rac,
        "localsecfilepath" => $localsecfilepath
    };

    $self->{'flag'} = "-" . $self->{'mode'};
    $self->{LocalarraySPNames} = \@LocalarraySPNames;
    $self->{mirnames} = \@mirnames;
    $self->{MIR_LUNS} = \%mir_luns;
    $self->{LUNS_OWNERSP} = \%lunowner_sp;
    $self->{Luns} = \@Luns;
    $self->{SSdisks} = \@SSdisks;
    $self->{dgname} = \@dgname;
    $self->{'SSOSHandles'} = ("");

    my $class = ref($self) || $self;
    bless $self;
    return $self;
}

sub DESTROY {}

sub _do {
    my ($do, $line,$ret);
    ($do) = (@_);
    VCSAG_LOG_MSG("N", "issuing command: $do", 101, "$do");
    open(INPUT,"$do 2>&1 |") || die("command $do failed");
    while ($line = <INPUT>) { chomp($line); $ret = ($ret) ? "$ret\n$line" : $line; }
    close INPUT;
    $_DOE=$?;
    $_DO = $ret;
    if ($_DOE) {
        my $ret = $_DOE / 256;
        system("$HALOG -add \"Command failed:\n$_DO\n\treturn code $ret\n\" -sev E");
    } else {
        VCSAG_LOG_MSG("N", "command succeeded", 102);
    }
    return $_DOE;
}

sub run_cmds {
    my $ret = 0;
    foreach my $cmd (@CMDS) {
        $ret = _do($cmd);
        if ($ret) { last; }
    }
    @CMDS = ();
    return $ret;
}

sub create_lockfile {
    my ($self) = @_;
    my $vcs_home=$VCSHOME;
    $vcs_home =~ s/\"//g;
    my $LockFile = $vcs_home . "/bin/MirrorViewSnap/.VCS_MirrorViewSnap_" . $self->{'resname'} . "_Online_Lockfile.DO-NOT-REMOVE";
    open(LFH, ">$LockFile");
    close(LFH);
}

sub remove_lockfile {
    my ($self) = @_;
    my $vcs_home=$VCSHOME;
    $vcs_home =~ s/\"//g;
    my $LockFile = $vcs_home . "/bin/MirrorViewSnap/.VCS_MirrorViewSnap_" . $self->{'resname'} . "_Online_Lockfile.DO-NOT-REMOVE";
    unlink("$LockFile");
}

sub lockfile_exists {
    my ($self) = @_;
    my $vcs_home=$VCSHOME;
    $vcs_home =~ s/\"//g;
    my $LockFile = $vcs_home . "/bin/MirrorViewSnap/.VCS_MirrorViewSnap_" . $self->{'resname'} . "_Online_Lockfile.DO-NOT-REMOVE";
    if ( -f "$LockFile") {
        return 1;
    } else {
        return 0;
    }
}

sub take_snapshot {
    my $self = shift;
    my $retVal = 0;
    my $luns = " ";
    my $lun = '';
    my $ownsp = '';

# Create a consistent and persistent session for all the luns
    my $i=0;
    for($i=0; $i <= $#{$self->{'mirnames'}}; $i++){
        my $mirror = $self->{mirnames}[$i];
        my $mir_lun = $self->{MIR_LUNS}{$mirror};

        $luns = $luns . $mir_lun . " ";
    }
    $retVal = $self->startSession($luns);
    if ($retVal) {
        return $retVal;
    }

    $i = 0;
    for($i=0; $i <= $#{$self->{'mirnames'}}; $i++){
        my $mirror = $self->{mirnames}[$i];
        my $mir_lun = $self->{MIR_LUNS}{$mirror};
        my $ownersp = $self->{LUNS_OWNERSP}{$mir_lun};

        my $snapshotName = "lun" . $mir_lun . "_" . "fd_snapshot";

        if ($^O !~ /^MSWin/i) {
            $self->validateSGandHost ($ownersp);
            $retVal = $self->validateLUN($ownersp, $mir_lun);
        }
        $retVal = $self->createSnapshot($ownersp, $mir_lun, $snapshotName);
        if ($retVal) {
            last;
        }
        $retVal = $self->activateSnapshot($ownersp, $snapshotName);
        if ($retVal) {
            last;
        }
        $retVal = $self->addSnapshotToSG($ownersp, $snapshotName);
        if ($retVal) {
            last;
        }
    }


# In case of an error in either -
#
# 1. creating a snapshot
# 2. activating a snapshot
# 3. adding a snapshot to a storage group
#
# roll back all the operations for all the mirrors even those for which we were successful in completing
# all the above operations.

    if ($retVal) {
        $i = ($i-1);
        for (; $i >= 0; $i--) {
            my $mirror = $self->{mirnames}[$i];
            my $mir_lun = $self->{MIR_LUNS}{$mirror};
            my $ownersp = $self->{LUNS_OWNERSP}{$mir_lun};

            my $snapshotName = "lun" . $mir_lun . "_" . "fd_snapshot";

            $self->deactivateSnapshot($ownersp, $snapshotName);
            $self->removeSnapshotFromSG($ownersp, $snapshotName);
            $self->removeSnapshot($ownersp, $snapshotName);
        }
        $self->stopSession();
    }
    return $retVal;
}

sub take_lun_snapshot {
    my $self = shift;
    my $retVal = 0;
    my $i=0;
    my $luns = '';
    my $ownsp = '';

# Create a consistent and persistent session for all the luns
    for($i=0; $i <= $#{$self->{'Luns'}}; $i++){
        $luns = $luns . $self->{'Luns'}[$i] . " ";
    }

    $retVal = $self->startSession($luns);
    if ($retVal) {
        return $retVal;
    }

    $i = 0;
    for($i=0; $i <= $#{$self->{'Luns'}}; $i++){
        my $lun = $self->{'Luns'}[$i];
        my $ownersp = $self->{LUNS_OWNERSP}{$lun};

        my $snapshotName = "lun" . $lun . "_" . "fd_snapshot";

        if ($^O !~ /^MSWin/i) {
            $self->validateSGandHost ($ownersp);
            $retVal = $self->validateLUN($ownersp, $lun);
        }
        $retVal = $self->createSnapshot($ownersp, $lun, $snapshotName);
        if ($retVal) {
            last;
        }
        $retVal = $self->activateSnapshot($ownersp, $snapshotName);
        if ($retVal) {
            last;
        }
        $retVal = $self->addSnapshotToSG($ownersp, $snapshotName);
        if ($retVal) {
            last;
        }
    }

# In case of an error in either -
#
# 1. creating a snapshot
# 2. activating a snapshot
# 3. adding a snapshot to a storage group
#
# roll back all the operations for all the mirrors even those for which we were successful in completing
# all the above operations.

    if ($retVal) {
        $i = ($i-1);
        for(; $i >= 0 ; $i--){
            my $lun = $self->{'Luns'}[$i];
            my $ownersp = $self->{LUNS_OWNERSP}{$lun};

            my $snapshotName = "lun" . $lun . "_" . "fd_snapshot";

            $self->deactivateSnapshot($ownersp, $snapshotName);
            $self->removeSnapshotFromSG($ownersp, $snapshotName);
            $self->removeSnapshot($ownersp, $snapshotName);
        }
        $self->stopSession();
    }
    return $retVal;
}

#
# Starting point from the online script
# sub pitc
# followed by sub ScanDev
#

sub pitc {
    my ($self) = @_;
    my $retVal = 0;

	if ($self->{'rac'} == 1) {
		# This is a RAC environment i.e. the group for which a 
		# FireDrill is being run is a parallel group.
		my $is_master = ArrayRep::is_cvm_master();

		if ($is_master == -1) {
			# An appropriate error message would
			# already have been logged in the
			# is_master sub
			return 1;
		}
		elsif ($is_master == 0) {
			# This node is the SLAVE
            my ($ot) = `ag_i18n_inc::HATYPE -value MirrorViewSnap OnlineTimeout`;
            chomp($ot);
            VCSAG_LOG_MSG("I", "This is a CVM Slave node. Online will now exit.", 45);
            exit ($ot);
		}
	}

	#
	# COME HERE ONLY IF THE NODE IS THE CVM MASTER OR THIS IS A
	# NON-RAC ENVIRONMENT
	#
    if ($^O !~ /^MSWin/i) {
        VCSAG_LOG_MSG("I", "\n\nVCS EMC MirrorViewSnap Agent\nPre-checks\n\n\n", 212);

        my $dmpcmd = "$VXDMPADM listenclosure all";
        my $dmpcmdout = `$dmpcmd`;
        VCSAG_LOG_MSG("I", "Environmental snapshot DMP pre-check online command:\n$dmpcmd\n\n$dmpcmdout\n", 213, "$dmpcmd", "$dmpcmdout");

        $self->ScanDev(2);
        $retVal = $self->ScanDev(3);
        if ($retVal) {
            VCSAG_LOG_MSG("W", "Unable to cleanup the EMC power path devices. Administrator intervention required. All the stale entries should be removed before online.", 144);
            return $retVal;
        }

        my $ddlcmd = "$VXDDLADM assign names";
        my $ddlcmdout = `$ddlcmd`;
        VCSAG_LOG_MSG("I", "PITC: Refreshing /etc/vx/disk.info: $ddlcmd\n$ddlcmdout\n", 166, "$ddlcmd", "$ddlcmdout");

    }

    if ($self->{'mode'}){
        $retVal = $self->take_snapshot();
    } else {
        $retVal = $self->take_lun_snapshot();
    }

    if ($retVal) {
        VCSAG_LOG_MSG("E", "Error taking snapshot. exiting", 104);
        return $retVal;
    } else {
        my (@fields) = split('\s+', `$HASYS -nodeid`);
        system("$HARES -modify $self->{'resname'} Responsibility $fields[5]");
    }

    if ($self->{'mode'}){
        $retVal = $self->mount_snapshot("gold");
    } else {
        $retVal = $self->mount_lun_snapshot();
    }

    return $retVal;
}

sub mount_snapshot {
    my ($self) = @_;

    my $ret = 0;
    my $i = 0;
    my $ddlcmd = "$VXDDLADM assign names";
    my (%vxdisk) = ();
    my ($tagid) = "gold-";

    if ($^O =~ /^MSWin/i) {
        $self->ScanDev();
        # Find the list of snapshots corresponding to the luns in the group/mirrors.
        my %luns;
        foreach my $mir (keys %{$self->{MIR_LUNS}}) {
            $luns{$self->{MIR_LUNS}{$mir}} = 1;
        }
        my $cmd = "$navisecpath -h $self->{'LocalarraySPNames'}[0] $self->{'localsecfileopt'} snapview -listsnapshots";
        my @snaps = split(/\n/, `$cmd`);
        my ($tmp_uid, $tmp_snaptarget, $tmp_pdev);
        my @snapuids1;
        foreach my $line (@snaps) {
            chomp $line;
            if ($line =~ /SnapView logical unit ID:\s+(\S+)/) {
                $tmp_uid = $1;
            }
            if ($line =~ /Target Logical Unit:\s+(\d+)/) {
                $tmp_snaptarget = $1;
                if (exists($luns{$tmp_snaptarget})) {
                    push(@snapuids1, $tmp_uid);
                }
            }
        }
        my $pat = join("|", @snapuids1);

        $cmd = "$VXDISK list";
        my @vxdisk_list = split(/\n/, `$cmd`);
        my @vxdisks;
        my $tagname = $tagid . $self->{'group'};
        foreach my $line (@vxdisk_list) {
            my @tokens = split(/\s+/, $line);
            my $num_tokens = $#tokens + 1;
            my $vdid = $tokens[$#tokens];
            my @match = grep(/$pat/, @tokens);
            my $nmatch = $#match + 1;

#next if ($pat !~ /$tokens[12]/i);
            next if ($nmatch <= 0);
            my $vxdisk = $tokens[0];
            my $cmd = "$VXDISK set tag value=$tagname $vxdisk";
            my $op = `$cmd`;
        }

        my $dgnamelist = join(' ', $self->{'dgname'});
        return $self->vxvm_test_diskgroup_import("gold", $tagname, 0);
    }
    VCSAG_LOG_MSG("I", "Scanning devices mount_snapshot:\n\n", 135);

    my $navi = "$navisecpath -h $self->{'LocalarraySPNames'}[$i] $self->{'localsecfileopt'}   storagegroup -list -gname $self->{'storageGrpName'} -host";
    my $logout = `$navi`;
    VCSAG_LOG_MSG("N", "\n\nEMC CLARiiON storagegroup (pre-dg-import) listing: $navi\n$logout\n", 167, "$navi", "$logout");

    VCSAG_LOG_MSG("N", "\n\nRefresh O/S device discovery\n", 101);

    $self->ScanDev(1);
    $self->ScanDev(0);

    my (@lines) = split('\n', `$VXDISK -eq list`);
    foreach my $l (@lines) {
        my (@fields) = split('\s+', $l);
        $vxdisk{$fields[6]} = $fields[0];
    }

    my $ddlcmdout = `$ddlcmd`;
    VCSAG_LOG_MSG("I", "Mount_snapshot: Refreshing /etc/vx/disk.info: $ddlcmd\n$ddlcmdout\n", 168, "$ddlcmd", "$ddlcmdout");

    my $dmpcmd = "$VXDMPADM listenclosure all";
    my $dmpcmdout = `$dmpcmd`;
    VCSAG_LOG_MSG("I", "Mount_snapshot: DMP enclosure details:\n$dmpcmd\n$dmpcmdout\n", 169, "$dmpcmd", "$dmpcmdout");

    my $vxdisklog = `$VXDISK -eo alldgs list`;
    VCSAG_LOG_MSG("I", "Mount_Snapshot: VM vxdisk output prior to dg imports:\n$vxdisklog\n\n", 170, "$vxdisklog");

    for($i=0; $i <= $#{$self->{'dgname'}}; $i++) {
        my (@STdisks) = ();
        my $dg = $self->{'dgname'}[$i];
        my $searchdg = $dg;
        my $newdg = $dg . "_fd";

        my $cmd = "$VXDISK -o alldgs list | $GREP -w udid_mismatch | $GREP -w '$searchdg'";
        my $cmdout = `$cmd`;
        VCSAG_LOG_MSG("I", "Mount_snapshot diskgroup udid_mismatch check command:\n$cmd\n$cmdout\n", 171, "$cmd", "$cmdout");

        my (@lines) = split('\n', $cmdout);
        foreach my $l (@lines) {
#VCSAG_LOG_MSG("I", "l: $l", 136, "$l");
            my (@fields) = split('\s+', $l);
            push(@STdisks, $fields[0]);
        }

        my $arrcount = scalar @STdisks;
        if ( $arrcount == 0 ) {
            my $cmd = "$VXDISK -o alldgs list | $GREP -w '$searchdg'";
            my $cmdout = `$cmd`;
            VCSAG_LOG_MSG("I", "Mount_snapshot diskgroup command:\n$cmd\n$cmdout\n", 172, "$cmd", "$cmdout");
            my (@lines) = split('\n', $cmdout);
            foreach my $l (@lines) {
#VCSAG_LOG_MSG("I", "l2: $l", 136, "$l");
                my (@fields) = split('\s+', $l);
                push(@STdisks, $fields[0]);
            }
        }

        my $tagname = $tagid . $self->{'group'};
        foreach my $disk (@STdisks) {
            push(@CMDS, "$VXDISK clearimport $disk");
            push(@CMDS, "$VXDISK settag $disk $tagname");
            $ret = run_cmds(@CMDS);
            if ($ret) {
                VCSAG_LOG_MSG("E", "cannot settag on disk $disk; exiting", 105, "$disk");
                $self->resynch(1);
                return 1;
            }
        }

        push(@CMDS, "$VXDG -C -n $newdg -o useclonedev=on -o updateid -o tag=$tagname import $dg");
        $ret = run_cmds(@CMDS);
        if ($ret) {
            $self->ScanDev(0);
            push(@CMDS, "$VXDG -C -n $newdg -o useclonedev=on -o updateid -o tag=$tagname import $dg");
            $ret = run_cmds(@CMDS);
            if ($ret) {
                VCSAG_LOG_MSG("E", "Mount_snapshot cannot import disk group $dg on snapshot; exiting", 173, "$dg");
                $self->resynch(1);
                return 1;
            }
        }

        $cmd = "$VXDISK -o alldgs list | $GREP -w '$dg'";
        $cmdout = `$cmd`;
        VCSAG_LOG_MSG("I", "\n\nMount_snapshot alldgs precheck command: $cmd\n$cmdout\n", 174, "$cmd", "$cmdout");

        push(@CMDS, "$VXDG deport $newdg");
        $ret = run_cmds(@CMDS);
        if ($ret) {
            VCSAG_LOG_MSG("E", "Mount_snapshot cannot deport disk group $newdg on snapshot; exiting", 175, "$newdg");
            return 1;
        }
    }
    return 0;
}

sub mount_lun_snapshot {
    my ($self) = @_;

    my $ret = 0;
    my (@STdisks);
    my (%vxdisk);
    my ($tagid) = "NonRepLocal-";
    my ($srcdisk) = 0;
    my ($i) = 0;

    if ($^O =~ /^MSWin/i) {
        $self->ScanDev();
        # Find the list of snapshots corresponding to the luns in the group/mirrors.
        my %luns;
        foreach my $mir (keys %{$self->{MIR_LUNS}}) {
            $luns{$self->{MIR_LUNS}{$mir}} = 1;
        }
        my $cmd = "$navisecpath -h $self->{'LocalarraySPNames'}[0] $self->{'localsecfileopt'} snapview -listsnapshots";
        my @snaps = split(/\n/, `$cmd`);
        my ($tmp_uid, $tmp_snaptarget, $tmp_pdev);
        my @snapuids1;
        foreach my $line (@snaps) {
            chomp $line;
            if ($line =~ /SnapView logical unit ID:\s+(\S+)/) {
                $tmp_uid = $1;
            }
            if ($line =~ /Target Logical Unit:\s+(\d+)/) {
                $tmp_snaptarget = $1;
                if (exists($luns{$tmp_snaptarget})) {
                    push(@snapuids1, $tmp_uid);
                }
            }
        }
        my $pat = join("|", @snapuids1);

        $cmd = "$VXDISK list";
        my @vxdisk_list = split(/\n/, `$cmd`);
        my @vxdisks;
        my $tagname = $tagid . $self->{'group'};
        foreach my $line (@vxdisk_list) {
            my @tokens = split(/\s+/, $line);
            my $num_tokens = $#tokens + 1;
            my $vdid = $tokens[$#tokens];
            my @match = grep(/$pat/, @tokens);
            my $nmatch = $#match + 1;

#next if ($pat !~ /$tokens[12]/i);
            next if ($nmatch <= 0);
            my $vxdisk = $tokens[0];
            my $cmd = "$VXDISK set tag value=$tagname $vxdisk";
            my $op = `$cmd`;
        }

        my $dgnamelist = join(' ', $self->{'dgname'});
        return $self->vxvm_test_diskgroup_import("gold", $tagname, 0);
    }
    $self->ScanDev(2);
    $self->ScanDev(0);

    my $ddlcmd = "$VXDDLADM assign names";
    my $ddlcmdout = `$ddlcmd`;
    VCSAG_LOG_MSG("I", "Mount_lun_snapshot: Refreshing /etc/vx/disk.info: $ddlcmd\n$ddlcmdout\n", 176, "$ddlcmd", "$ddlcmdout");

    my $dmpcmd = "$VXDMPADM listenclosure all";
    my $dmpcmdout = `$dmpcmd`;
    VCSAG_LOG_MSG("I", "Mount_lun_snapshot: DMP enclosure details:\n$dmpcmd\n$dmpcmdout\n", 177, "$dmpcmd", "$dmpcmdout");

    for($i=0; $i <= $#{$self->{'dgname'}}; $i++){
        my $dg = $self->{'dgname'}[$i];
        my @lines = split('\n', `$VXDISK -o alldgs list | $GREP -w udid_mismatch | $GREP -w '$dg'`);
        foreach my $l (@lines) {
            $srcdisk = 0;
            my (@fields) = split('\s+', $l);

            for($i=0; $i <= $#{$self->{'SSdisks'}}; $i++){
                my $disk = $self->{'SSdisks'}[$i];
                if ($disk ne $fields[0]){ next; }
                $srcdisk = 1;
            }
            if ($srcdisk){
                next;
            }
            push(@STdisks, $fields[0]);
        }

        @lines = split('\n', `$VXDISK -eq list`);
        foreach my $l (@lines) {
            my (@fields) = split('\s+', $l);
            $vxdisk{$fields[$#fields]} = $fields[0];
        }

# TBD unique tagname across clusters
        my $tagname = $tagid . $self->{'group'};
        foreach my $disk (@STdisks) {
            push(@CMDS, "$VXDISK clearimport $disk");
            push(@CMDS, "$VXDISK settag $disk $tagname");
            $ret = run_cmds(@CMDS);
            if ($ret) {
                $self->resynch(1);
                return 1;
            }
        }

        my $newdg = $dg . "_fd";
        push(@CMDS, "$VXDG -C -n $newdg -o useclonedev=on -o updateid -o tag=$tagname import $dg");
        $ret = run_cmds(@CMDS);
        if ($ret) {
            VCSAG_LOG_MSG("E", "cannot import disk group $dg on snapshot; exiting", 105, "$dg");
            $self->resynch(1);
            return 1;
        }

        my $cmd = "$VXDISK -o alldgs list | $GREP -w '$dg'";
        my $cmdout = `$cmd`;
        VCSAG_LOG_MSG("I", "\n\nMount_lun_snapshot alldgs precheck command: $cmd\n$cmdout\n", 178, "$cmd", "$cmdout");

        push(@CMDS, "$VXDG deport $newdg");
        $ret = run_cmds(@CMDS);
        if ($ret) {
            VCSAG_LOG_MSG("E", "cannot deport disk group $newdg on snapshot; exiting", 106, "$newdg");
            return 1;
        }
    }
    return 0;
}

sub cleanup_aix_devices
{
    my ($self) = shift;
    foreach my $oshandle (split('\n', $self->{'SSOSHandles'})) {
        chomp $oshandle;
        next if (! $oshandle);
        my $cmd = "/usr/sbin/rmdev -dl $oshandle";
        push(@CMDS, $cmd);
    }
    push(@CMDS, $CFGMGR);
    run_cmds(@CMDS);
    return 0;
}

sub cleanup_linux_devices
{
    my ($self) = shift;
    my ($kernel_release) = split('\n', `uname -r`);
    if ($kernel_release =~ /2.6/) {
        my @oshandles = split('\n', $self->{'SSOSHandles'});
        foreach my $oshandle (split('\n', $self->{'SSOSHandles'})) {
            chomp $oshandle;
            next if (! $oshandle);
            my $cmd = "echo 1 > /sys/block/$oshandle/device/delete";
            VCSAG_LOG_MSG("I", "Deleting device [$oshandle] using cmd [$cmd]", 150, "$oshandle", "$cmd");
            my $op = `$cmd`;
        }
    }
    VCSAG_LOG_MSG("I", "Rescaning VxVM devices", 160);
    push(@CMDS, "$VXDISK scandisks");
    my $ret = run_cmds(@CMDS);
    if ($ret){
        VCSAG_LOG_MSG("E", "VxVM Device scanning failed. Hence exiting.", 161);
        exit 1;
    }
    if (-f "/usr/bin/lsscsi") {
        my $linscan = "/usr/bin/lsscsi | $GREP DGC";
        my $linlog = `$linscan`;
        VCSAG_LOG_MSG("I", "\n\nLinux device tree output :$linscan\n$linlog\n", 162, "$linscan", "$linlog");
    }
    return 0;
}

sub remove_lun_mappings {
    my ($self) = shift;
    my $PP = shift;

    my $ret = 0;
    my $i = 0;

    if ($SYSTEM eq "Linux") {
        return $self->cleanup_linux_devices();
    }
    if ($SYSTEM eq "AIX") {
        return $self->cleanup_aix_devices();
    }
    if ($SYSTEM ne "SunOS") {
        return 0;
    }

    if ($PP) {
        $ret = $self->cleanup_cfg_stack();
        VCSAG_LOG_MSG("I", "Updating powerpath data after removing lun mapping\n\n", 135);
        push(@CMDS, "$DEVFSADM -Cvc disk");


        my $ppcmd = "$POWERCF -q";
        my $pplog = `$ppcmd`;
        VCSAG_LOG_MSG("I", "PP remove_lun_mappings command ouput :\n$ppcmd\n$pplog\n", 179, "$ppcmd", "$pplog");

        $ppcmd = "$POWERMT config";
        VCSAG_LOG_MSG("I", "PP remove_lun_mappings command: $ppcmd", 180, "$ppcmd");
        $ret = system($ppcmd);

        $ppcmd = "$POWERMT save";
        VCSAG_LOG_MSG("I", "PP remove_lun_mappings command: $ppcmd", 181, "$ppcmd");
        $ret = system($ppcmd);

        $ppcmd = "$POWERMT display";
        my $ppcmdout_for_log = `$ppcmd`;
        VCSAG_LOG_MSG("I", "PP remove_lun_mappings LUN Count: $ppcmd\n$ppcmdout_for_log\n", 138, "$ppcmd", "$ppcmdout_for_log");


        push(@CMDS, "$VXDISK scandisks");
#       push(@CMDS, "$VXDDLADM assign names");
        my $ret = run_cmds(@CMDS);
        if ($ret){
            VCSAG_LOG_MSG("E", "Device scanning using EMC PowerPath (powermt) failed. Hence exiting.", 182);
            exit 1;
        }
    } else {
        VCSAG_LOG_MSG("I", "Updating VxVM data after removing lun mapping", 135);
        push(@CMDS, "$DEVFSADM -Cvc disk");
        push(@CMDS, "$VXDISK scandisks");
#        push(@CMDS, "$VXDDLADM assign names");

        $ret = run_cmds(@CMDS);
        if ($ret) {
            VCSAG_LOG_MSG("E", "Updating VxVM data after removing lun mapping failed", 135);
            return 2;
        }
    }

    return 0;
}

sub cleanup_cfg_stack {
    my ($self) = @_;

    my $ret = 0;
    my $i = 0;
    my $currapid = "";
    my $prevapid = "";
    my (@apidarr) = ();

    my $cmd = "$CFGADM -alo show_FCP_dev";
    VCSAG_LOG_MSG("I", "Issuing command: $cmd", 136, "$cmd");
    my @cfgcmdout = split('\n', `$cmd`);

    foreach my $cmdline (@cfgcmdout) {
        if ($cmdline =~ /Ap_Id/) {
            next;
        }

        if ($cmdline =~ /unusable/) {
            my @cmdarr = split('\s+', $cmdline);
            if ( $cmdarr[4] eq "unusable" ) {
                $currapid = $cmdarr[0];
                ($currapid, undef) = split(',', $currapid);

                if ( $currapid ne $prevapid ) {
                    push(@apidarr, $currapid);
                }

                $prevapid = $currapid;
            }
        } elsif ($cmdline =~ /failing/) {
            my @cmdarr = split('\s+', $cmdline);
            if ( $cmdarr[4] eq "failing" ) {
                my $apid = $cmdarr[0];
                my $rdsk = "/dev/rdsk/";
                my @arr = split(':', $apid);
                $rdsk = $rdsk . $arr[0];
                chomp($rdsk);
                @arr = split(',', $arr[2]);
                my $id = uc $arr[0];
                chomp($id);
                $rdsk = $rdsk . "t" . $id . "d" . $arr[1] . "s2";
                chomp($rdsk);
                my $luxadmcmd = "$LUXADM -e offline $rdsk";
                VCSAG_LOG_MSG("W", "Condition for Ap_Id $apid is failing. Hence running luxadm command to make it unusable", 134, "$apid");
                push(@CMDS, $luxadmcmd);
                my $ret = run_cmds(@CMDS);
                $ret = $ret / 256;
                if ($ret){
                    VCSAG_LOG_MSG("W", "Luxadm command failed.", 138);
                    return $ret;
                }

                my ($currapid, undef) = split(',', $apid);
                push(@apidarr, $currapid);
            }
        }
    }

    foreach my $currapid (@apidarr) {
        $cmd = "$CFGADM -o unusable_FCP_dev -c unconfigure $currapid 2>&1 >/dev/null";
        VCSAG_LOG_MSG("I", "Issuing command: $cmd", 136, "$cmd");
        @cfgcmdout = split('\n', `$cmd`);
        $ret = $?;
        if ($ret) {
            VCSAG_LOG_MSG("W", "Unable to unconfigure apid $currapid", 134, 
                    "$currapid");
            return $ret;
        }
    }

    return 0;
}

sub restore_emc_stack {
    my ($self) = @_;
    my $ret = 0;


    my $ppcmd = "$POWERCF -q";
    my $pplog = `$ppcmd`;
    VCSAG_LOG_MSG("I", "PP restore_emc_stack command ouput :\n$ppcmd\n$pplog\n", 183, "$ppcmd", "$pplog");


    $ppcmd = "yes | $POWERMT restore dev=all";
    my $ppcmdout_for_log = `$ppcmd`;
    VCSAG_LOG_MSG("I", "PP restore_emc_stack command: $ppcmd\n$ppcmdout_for_log\n", 184, "$ppcmd", "$ppcmdout_for_log");

    $ppcmd = "$POWERMT check dev=all class=clariion force";
    $ppcmdout_for_log = `$ppcmd`;
    VCSAG_LOG_MSG("I", "PP restore_emc_stack command: $ppcmd\n$ppcmdout_for_log\n", 185, "$ppcmd", "$ppcmdout_for_log");

    $ppcmd = "$POWERCF -q";
    $pplog = `$ppcmd`;
    VCSAG_LOG_MSG("I", "PP restore_emc_stack command :\n$ppcmd\n$pplog\n", 186, "$ppcmd", "$pplog");

    $ppcmd = "$POWERMT display";
    $ppcmdout_for_log = `$ppcmd`;
    VCSAG_LOG_MSG("I", "PP restore_emc_stack LUN Count: $ppcmd\n$ppcmdout_for_log\n", 138, "$ppcmd", "$ppcmdout_for_log");

    my $vmcmd = "$VXDISK scandisks";
    VCSAG_LOG_MSG("I", "VM restore_emc_stack command: $vmcmd", 187, "$vmcmd");

    my $dmpcmd = "$VXDMPADM listenclosure all";
    my $dmpcmdout = `$dmpcmd`;
    VCSAG_LOG_MSG("I", "DMP restore_emc_stack LUN Count:\n$dmpcmd\n$dmpcmdout\n", 188, "$dmpcmd", "$dmpcmdout");

    my $ddlcmd = "$VXDDLADM assign names";
    my $ddlcmdout = `$ddlcmd`;
    VCSAG_LOG_MSG("I", "Refreshing /etc/vx/disk.info: $ddlcmd\n$ddlcmdout\n", 189, "$ddlcmd", "$ddlcmdout");

    return 0;
}

sub remove_powerpath_disks {
    my ($self) = @_;
    my $ret = 0;
    my $i = 0;
    my $diskrmcmd = "$VXDISK rm";
    my $ppdiskrmcmd = "$POWERMT remove device=";
    my $disklist = " ";
    my $ppdisklist = "";

    my (@VMdisks) = ();
    my (@PPdisks) = ();
    my (@vxdiskout) = ();
    my (@vxdiskeqout) = ();

    VCSAG_LOG_MSG("I", "Removing EMC PowerPath disks created after snapshot.", 133);
    my @val = split ('\n', `$VXDISK -o alldgs list`);
    push (@vxdiskout, @val);
    $ret = $?;
    if ($ret) {
        return  1;
    }

    for($i=0; $i <= $#{$self->{'dgname'}}; $i++){
        my $dg = $self->{'dgname'}[$i];
        my $snapdg = $dg . "_fd";

        foreach my $diskline (@vxdiskout) {
            if ( $diskline =~ /DEVICE/ ) {
                next;
            }

            if ( ($diskline =~ /\b$snapdg\b/) || ($diskline =~ /\b$dg\b/) ) {
                my @diskarr = split('\s+', $diskline);
                my $disk = $diskarr[0];
                push(@VMdisks, $disk);
                $disklist = $disklist . $disk . " ";
            }
        }
    }

    if ( $disklist eq " ") {
        VCSAG_LOG_MSG("W", "Disk list is empty, hence can't remove.", 138);
        return 1;
    }

# Cache the OS Device names on linux and AIX for deleting later.
    if (($SYSTEM eq "Linux") || ($SYSTEM eq "AIX")) {
        foreach my $vmdisk (split(' ', $disklist)) {
                my $cmd = "$POWERMT display dev=$vmdisk | $AWK \'\/hdisk[0-9]\/{print \$3}\'";
                my @hdisklist = `$cmd`;
                chomp @hdisklist;
                foreach my $osdisk (@hdisklist) {
                    $self->{'SSOSHandles'} = $self->{'SSOSHandles'} . "\n" . $osdisk;
                }

        }
    }


    @val = split ('\n', `$VXDISK -eq list`);
    push (@vxdiskeqout, @val);

    foreach my $vmdisk (@VMdisks) {
        foreach my $line (@vxdiskeqout) {
            if ($line =~ /\b$vmdisk\b/){
                my (@arr) = split('\s+', $line);
                if($line =~ /clone_disk/ || $line =~ /udid_mismatch/){
                   $ppdisklist = $ppdisklist . $arr[6] . ",";
                   push(@PPdisks, $arr[6]);
                } else {
                   $ppdisklist = $ppdisklist . $arr[5] . ",";
                   push(@PPdisks, $arr[5]);
                }
                last;
            }
        }
    }

    chop $ppdisklist;

    $diskrmcmd = $diskrmcmd . $disklist;
    push(@CMDS, "$diskrmcmd");
    $ret = run_cmds(@CMDS);
    if ($ret) {
        VCSAG_LOG_MSG("W", "Failed to remove the disks $disklist", 132, "$disklist");
        return 1;
    }


    foreach my $ppdisk (@PPdisks) {
        my $ppcmd = $ppdiskrmcmd . $ppdisk;
        push(@CMDS, $ppcmd);
        $ret = run_cmds(@CMDS);
        if ($ret) {
            sleep 1;
            $ret = run_cmds(@CMDS);
            if ( $ret ) {
                VCSAG_LOG_MSG("N", "Removing PowerPath disk $ppdisk failed", 142, "$ppdisk");
            }
        }
    }

    return $ret;
}

sub remove_vmdisks {
    my ($self) = shift;
    my $PP = shift;

    my $ret = 0;
    my $i = 0;
    my $diskrmcmd = "$VXDISK rm";
    my $disklist = " ";

    if (($SYSTEM ne "Linux") && ($SYSTEM ne "SunOS") && ($SYSTEM ne "AIX")) {
        return 0;
    }
    if ($PP) {
        $ret = $self->remove_powerpath_disks();
        return $ret;
    }

    my @vxdiskout =  split ('\n', `$VXDISK -eo alldgs list`);
    $ret = $?;
    if ($ret) {
        return  1;
    }

    for($i=0; $i <= $#{$self->{'dgname'}}; $i++){
        my $dg = $self->{'dgname'}[$i];
        my $snapdg = $dg . "_fd";

        foreach my $diskline (@vxdiskout) {
            if ( $diskline =~ /DEVICE/ ) {
                next;
            }

            if ( ($diskline =~ /\b$snapdg\b/) || ($diskline =~ /\b$dg\b/) ) {
                my @diskarr = split('\s+', $diskline);
                my $disk = $diskarr[0];
                $disklist = $disklist . $disk . " ";
            }
        }
    }
    if ( $disklist eq " ") {
        VCSAG_LOG_MSG("W", "Disk list is empty, hence can't remove.", 138);
        return 0;
    }

# Cache the OS Device names on linux and AIX for deleting later.
    if (($SYSTEM eq "Linux") || ($SYSTEM eq "AIX")) {
        foreach my $vmdisk (split(' ', $disklist)) {
            my $cmd = "$VXDMPADM getsubpaths dmpnodename=$vmdisk | $AWK \'\/EMC_CLARiiON\/{print \$1}\'";
            VCSAG_LOG_MSG("I", "Issuing command [$cmd] to find subpaths.", 151, "$cmd");
            my $op = `$cmd`;
            VCSAG_LOG_MSG("I", "Output: [$op]", 152, "$op");
            $self->{'SSOSHandles'} = $self->{'SSOSHandles'} . "\n" . $op;
        }
    }

    $diskrmcmd = $diskrmcmd . $disklist;
    VCSAG_LOG_MSG("I", "Issuing command to remove disks: $diskrmcmd", 131, "$diskrmcmd");
    push(@CMDS, "$diskrmcmd");
    $ret = run_cmds(@CMDS);
    if ($ret) {
        VCSAG_LOG_MSG("W", "Failed to remove the disks $disklist", 132, "$disklist");
        return 1;
    }
 
    $ret = run_cmds(@CMDS);
    if ($ret) {
        VCSAG_LOG_MSG("N", "Disabling VxVM DMP path for disks $disklist failed", 141, "$disklist");
    }

}


sub resynch {
    my ($self) = shift;
    my $rmflag = shift;
    my $retVal = 0;
    my $cmd = '';
    my $i = 0;
    my $lun = '';
    my $ownersp = '';

    my ($dmpcmd, $dmpcmdout, $vxdisklog);
    if ($^O !~ /^MSWin/i) {
        VCSAG_LOG_MSG("I", "\n\n\n\nVCS EMC MirrorViewSnap Agent\noffline-checks\n\n\n", 190);

        $dmpcmd = "$VXDMPADM listenclosure all";
        $dmpcmdout = `$dmpcmd`;
        VCSAG_LOG_MSG("I", "Environmental snapshot DMP offline-checks command:\n$dmpcmd\n\n$dmpcmdout\n", 191, "$dmpcmd", "$dmpcmdout");

        $vxdisklog = `$VXDISK -eo alldgs list`;
        VCSAG_LOG_MSG("I", "VM vxdisk -eq output prior to offline:\n$vxdisklog\n\n", 192, "$vxdisklog");
    }
    my $PP = is_emc_powerpath_enable();
    my ($ppcmd, $ppcmdout_for_log, $pplog);
    if ($PP) {
        $ppcmd = "$POWERMT display";
        $ppcmdout_for_log = `$ppcmd`;
        VCSAG_LOG_MSG("I", "PP LUN Count prior to LUN removal offline: $ppcmd\n$ppcmdout_for_log\n", 193, "$ppcmd", "$ppcmdout_for_log");
    }

    VCSAG_LOG_MSG("I", "Removing VxVM disks created after snapshot.\n\n", 133);
    $retVal = $self->remove_vmdisks($PP);
    if ($retVal) {
        VCSAG_LOG_MSG("E", "Removing VxVM disks created after snapshot failed. Hence exiting.", 135);
        return $retVal;
    }


    if ($self->{'mode'}) {
        for($i=0; $i <= $#{$self->{'mirnames'}}; $i++){
            my $mirror = $self->{mirnames}[$i];
            my $mir_lun = $self->{MIR_LUNS}{$mirror};
            my $ownersp = $self->{LUNS_OWNERSP}{$mir_lun};

            my $snapshotName = "lun" . $mir_lun . "_" . "fd_snapshot";
            my $sessionName = "lun" . $mir_lun . "_" . "fd_session";

            $retVal = $self->deactivateSnapshot($ownersp, $snapshotName);
            $retVal = $self->removeSnapshotFromSG($ownersp, $snapshotName);
            $retVal = $self->removeSnapshot($ownersp, $snapshotName);
        }
    } else {
        for($i=0; $i <= $#{$self->{'Luns'}}; $i++){
            my $lun = $self->{'Luns'}[$i];
            my $ownersp = $self->{LUNS_OWNERSP}{$lun};

            my $snapshotName = "lun" . $lun . "_" . "fd_snapshot";
            my $sessionName = "lun" . $lun . "_" . "fd_session";
            $retVal = $self->deactivateSnapshot($ownersp, $snapshotName);
            $retVal = $self->removeSnapshotFromSG($ownersp, $snapshotName);
            $retVal = $self->removeSnapshot($ownersp, $snapshotName);
        }
    }

    $retVal = $self->stopSession();
    my (@fields) = split('\s+', `$HASYS -nodeid`);
    system("$HARES -modify $self->{'resname'} Responsibility \"\"");


    if ($^O =~ /^MSWin/i) {
        $self->ScanDev();
        return $retVal;
    }

    if ($PP) {
        $ppcmd = "$POWERCF -q";
        $pplog = `$ppcmd`;
        VCSAG_LOG_MSG("I", "PP clean-up $ppcmd during offline: \n$pplog\n", 194, "$ppcmd", "$pplog");
    }


#   if ( $rmflag != 1 ) {
    VCSAG_LOG_MSG("I", "Removing FC LUN mapping from server.\n\n", 134);
    $retVal = $self->remove_lun_mappings($PP);
    if ( $retVal == 1 ) {
        $self->ScanDev(1); # For other platform, we need to do scandev
    }
#        }


    # Rescan the vxvm disks so that the snap devs don't appear after assign names.
    my $vxdiskscancmd = "$VXDISK scandisks";
    my $op = `$vxdiskscancmd`;

    my $ddlcmd = "$VXDDLADM assign names";
    my $ddlcmdout = `$ddlcmd`;
    VCSAG_LOG_MSG("I", "Refreshing /etc/vx/disk.info: $ddlcmd\n$ddlcmdout\n", 196, "$ddlcmd", "$ddlcmdout");

    VCSAG_LOG_MSG("I", "Environmental snapshot DMP at the end of offline process: \n", 131);
    $dmpcmd = "$VXDMPADM listenclosure all";
    $dmpcmdout = `$dmpcmd`;
    VCSAG_LOG_MSG("I", "DMP LUN Count: $dmpcmd\n$dmpcmdout\n", 195, "$dmpcmd", "$dmpcmdout");

    if ($PP) {
        $ppcmd = "$POWERMT display";
        $ppcmdout_for_log = `$ppcmd`;
        VCSAG_LOG_MSG("I", "PP LUN Count: $ppcmd\n$ppcmdout_for_log\n", 197, "$ppcmd", "$ppcmdout_for_log");
    }

    return $retVal;
}

sub do_command {
    my ($cmd) = @_;
    my $output = `$cmd 2>&1`;
    my $status = $? >> 8;

    return ($status, $output);
}


sub getlunowner {
    my($numSPs) = shift;
    my(@LocalarraySPNames) = '';

    my $i = 0;
    for($i=0; $i<=$numSPs; $i++){
        $LocalarraySPNames[$i] = shift;
    }
    my($lun) = shift;
    my ($localsecfileopt) = shift;

    my $lunowner = '';
    my $ownersp = '';

    my $q;
    my $cmd = "$navisecpath -h $LocalarraySPNames[0] $localsecfileopt getlun $lun -snapshot -owner";
    my @query = split('\n', `$cmd`);
    foreach $q (@query) {
        if ($q =~ "Invalid LUN number"){
            VCSAG_LOG_MSG("N", "Could not find the specified lun : $lun. Exiting ...\n", 108, $lun);
            exit 1;
        }

        if ($q =~ /Current owner/) {
            my @owner = split('\s+', $q);
            $lunowner = $owner[3];
            last;
        }
    }

# Get owner Storage processor for lun 

    my %sp_serial = ();
    my $storage_proc = '';
    foreach my $sp (@LocalarraySPNames){
        $cmd = "$navisecpath -h $sp $localsecfileopt getall -sp";
        @query = split('\n', `$cmd`);
        foreach $q (@query) {
            if ($q =~ /^Storage Processor:\s+SP (\S+)/){
                $storage_proc = $1;
                next;
            }

            if ($q =~ /^Serial Number For The SP:\s+(\S+)/){
                if ($1 eq "N/A"){
                    next;
                } 

                if ($storage_proc eq $lunowner){
                    $ownersp = $sp;
                    return $ownersp;
                }
            }
        }
    }

    return $ownersp;
}

sub deactivateSnapshot {
    my $self = shift;
    my($ownersp) = shift;
    my($snapshotName) = shift;
    my $retVal = 0;
    VCSAG_LOG_MSG("N", "Deactivating snapshot : $snapshotName.\n", 109, $snapshotName);

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} snapview -deactivatesnapshot -snapshotname $snapshotName -o");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("N", "Error trying to deactivate snapshot : $snapshotName\n", 110, $snapshotName);
    }
    return $retVal;
}

sub stopSession {
    my $self = shift;

    my $retVal = 1;
    my $i = 0;
    VCSAG_LOG_MSG("N", "Stopping session : $self->{'sessionName'}\n", 111, $self->{'sessionName'});

    for($i=0; $i <= $#{$self->{'LocalarraySPNames'}}; $i++){
        my $cmd = "$navisecpath -h $self->{'LocalarraySPNames'}[$i] $self->{'localsecfileopt'} snapview -stopsession $self->{'sessionName'} -o 2>&1";
        VCSAG_LOG_MSG("N", "issuing command: $cmd", 101, "$cmd");
        my @query = split('\n', `$cmd`);
        foreach my $q (@query) {
            if ($q =~ /This request must be issued from the SP that owns the session/){
                my $session_name  = $self->{'sessionName'};
                my $localspname = $self->{'LocalarraySPNames'}[$i];
                VCSAG_LOG_MSG("E", "Error while trying to stop session : $session_name. $localspname is not the owner of the session.\n", 130, $session_name, $localspname);
                last;
            }
        }
        $retVal = 0;

        $cmd = "$navisecpath -h $self->{'LocalarraySPNames'}[$i] $self->{'localsecfileopt'} storagegroup -list -gname $self->{'storageGrpName'} -host";
        my $logout = `$cmd`;
        VCSAG_LOG_MSG("N", "\n\nEMC CLARiiON storagegroup listing: $cmd\n$logout\n", 101, "$cmd", "$logout");
    }

    return $retVal;
}

sub removeSnapshotFromSG {
    my $self = shift;
    my($ownersp) = shift;
    my($snapshotName) = shift;
    my $retVal = 0;

    VCSAG_LOG_MSG("N", "Removing snapshot : $snapshotName from Storage group : $self->{'storageGrpName'}\n", 113, $snapshotName, $self->{'storageGrpName'});

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} storagegroup -removesnapshot -gname $self->{'storageGrpName'} -snapshotname $snapshotName -o");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("N", "Error trying to remove the snapshot $snapshotName from the storage group : $self->{'storageGrpName'}\n", 114, $snapshotName, $self->{'storageGrpName'});
    }
    return $retVal;
}

sub removeSnapshot {
    my $self = shift;
    my($ownersp) = shift;
    my($snapshotName) = shift;
    my $retVal = 0;
    VCSAG_LOG_MSG("N", "Removing Snapshot : $snapshotName\n", 115, $snapshotName);

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} snapview -rmsnapshot -snapshotname $snapshotName -o");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("N", "Error trying to remove the snapshot : $snapshotName\n", 116, $snapshotName);
    }
    return $retVal;
}



sub createSnapshot {
    my $self = shift;
    my($ownersp) = shift;
    my ($mir_lun) = shift;
    my ($snapshotName) = shift;
    my $retVal = 0;
    VCSAG_LOG_MSG("N", "Creating Snapshot : $snapshotName\n", 117, $snapshotName);

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} snapview -createsnapshot $mir_lun -snapshotname $snapshotName");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("N", "Error trying to create snapshot : $snapshotName\n", 118, $snapshotName);
    }

    return $retVal;
}

sub startSession {
    my $self = shift;
    my($luns) = shift;
    my($retVal) = 0;

    VCSAG_LOG_MSG("I", "\n\n\n\nEMC CLARiiON STORAGE COMMANDS.\n\n", 101);

    VCSAG_LOG_MSG("N", "Starting session : $self->{'sessionName'}\n", 119, $self->{'sessionName'});

    push(@CMDS, "$navisecpath -h $self->{LocalarraySPNames}[0] $self->{'localsecfileopt'} snapview -startsession $self->{'sessionName'} -lun $luns -persistence -consistent");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("E", "Error trying to start a session : $self->{'sessionName'}.\n", 120, $self->{'sessionName'});
    }

    return $retVal;
}

sub activateSnapshot {
    my $self = shift;
    my($ownersp) = shift;
    my($snapshotName) = shift;
    my $retVal = 0;
    VCSAG_LOG_MSG("N", "Activating snapshot : $snapshotName.\n", 121, $snapshotName);

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} snapview -activatesnapshot $self->{'sessionName'} -snapshotname $snapshotName");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("E", "Error trying to activate snapshot : $snapshotName.\n", 122, $snapshotName);
    }

    return $retVal;
}

sub addSnapshotToSG {
    my $self = shift;
    my($ownersp) = shift;
    my ($snapshotName) = shift;
    my $retVal = 0;
    VCSAG_LOG_MSG("N", "Adding snapshot : $snapshotName to Storage group : $self->{'storageGrpName'}\n", 123, $snapshotName, $self->{'storageGrpName'});

    push(@CMDS, "$navisecpath -h $ownersp $self->{'localsecfileopt'} storagegroup -addsnapshot -gname $self->{'storageGrpName'} -snapshotname $snapshotName");
    $retVal = run_cmds(@CMDS);
    $retVal = $retVal / 256;
    if ($retVal){
        VCSAG_LOG_MSG("E", "Error trying to add the snapshot '$snapshotName' to the storage group : $self->{'storageGrpName'}\n", 124, $snapshotName, $self->{'storageGrpName'});
    }

    return $retVal;
}


sub validateSGandHost {
    my $self = shift;
    my($ownersp) = shift;
    my (@hosts) = '';
    my $hostvalid_yn = "n";

    my $cmd = "$navisecpath -h $ownersp $self->{'localsecfileopt'} storagegroup -list -gname $self->{'storageGrpName'} -host";
    my $logout = `$cmd`;
    VCSAG_LOG_MSG("N", "\n\nEMC CLARiiON storagegroup listing: $cmd\n$logout\n", 101, "$cmd", "$logout");

    my @query = split('\n', $logout);
    foreach my $q (@query) {

        if ($q =~ /The group name or UID does not match any storage groups for this array/){
            VCSAG_LOG_MSG("E", "Could not find Storage Group : $self->{'storageGrpName'}.\n", 125, $self->{'storageGrpName'});
            exit 1;
        }
        if ($q !~ /Host name/) {
            next;
        }

        my @hostname = split('\s+', $q);
        push(@hosts, $hostname[2]);
    }

    if ($#hosts < 1){
        VCSAG_LOG_MSG("E", "No hosts found that are bound to storage group $self->{'storageGrpName'}\n", 126, $self->{'storageGrpName'});
        exit 1;
    }

    my ($localhost) = $nodename;
    my $binAddr = Socket::inet_aton ($localhost);
    my $dotAddr = Socket::inet_ntoa ($binAddr);
    foreach my $h (@hosts){
        if($h =~/^[0-9]/){
            if ($dotAddr != $h){
                next;
            }
        } else {
            if ($localhost ne $h){
                $localhost = gethostbyaddr($binAddr, AF_INET);
                if ($localhost ne $h){
                    my ($host) = split('\.', $h);
                    next if (!defined($host) || ($host ne $localhost));
                }
            }
        }

        $hostvalid_yn = "y";
        last;
    }

    if ($hostvalid_yn ne "y") {
        VCSAG_LOG_MSG("E", "The host : $localhost is not bound to the storage group $self->{'storageGrpName'}. Cannot run fire-drill on $localhost. \n", 127, $localhost, $self->{'storageGrpName'}, $localhost); 
        exit 1;
    }
}

sub validateLUN {
    my $self = shift;
    my($ownersp) = shift;
    my($mir_lun) = shift;
    my $retVal;

    my $cmd = "$navisecpath -h $ownersp $self->{'localsecfileopt'} reserved -lunpool -list -freeluns";
    my @query = split('\n', `$cmd`);
    foreach my $q (@query) {

        if ($q !~ /^Unallocated LUNs/){
            next;
        }

        my @freeluns = split('\s+', $q);

        if ($#freeluns < 2) {
            VCSAG_LOG_MSG("E", "Not enough luns in the reserved lun pool. Please add luns in the reserved lun pool for snapshot to succeed.\n", 128);
            exit 1;
        }
    }

    $cmd = "$navisecpath -h $ownersp $self->{'localsecfileopt'} getlun $mir_lun -name | sed 's/^Name[ \t]*//' | xargs";
    my $mir_lun_name = `$cmd`;

    $cmd = "$navisecpath -h $ownersp $self->{'localsecfileopt'} snapview -listsnapableluns";
    my $cmdout = `$cmd`;
    $retVal = $?;
    $retVal = $retVal/256;
    if ($retVal){
        VCSAG_LOG_MSG("E", "Lun $mir_lun is not snappable \n", 129, $mir_lun);
        exit 1;
    }

    if ( $cmdout =~ /$mir_lun_name/ ) {
        $retVal = 0;
    } else {
        $retVal = 1;
    }

    return $retVal;
}


#
# is_emc_powerpath_enable - it tells whether the EMC PowerPath is enabled or not
#
sub is_emc_powerpath_enable {
    my $PP = 0;
    my $cmd;
    if ($SYSTEM eq "SunOS") {
        $cmd = $MODINFO;
    }elsif ($SYSTEM eq "AIX"){
             $cmd = $GENKEX; 
    }else{
          return 0;
    }
    my (@modinfoout) = split('\n', `$cmd`);

    if ($SYSTEM eq "SunOS") {
        foreach my $line (@modinfoout) {
           if ( ($line =~ /emcp/) && ($line =~ /PP Driver/) ) {
                $PP = 1;
                last;
           }
        }
    } elsif ($SYSTEM eq "AIX") {
        foreach my $line (@modinfoout) {
            if ( ($line =~ /powerdiskdd/) ){
                $PP = 1;
                last;
            }
        }
    }
    my $dmp = "vxdmpadm gettune all | $GREP dmp_monitor_osevent";

    my (@dmpout) = split(' ', `$dmp`);


    my $ppcmd;
    if ($dmpout[0] ne "dmp_monitor_osevent" && $PP == 1 )  {
        $ppcmd = "$VXDDLADM stop eventsource";
        VCSAG_LOG_MSG("I", "EMC PowerPath detected, stopping eventsource:\n$ppcmd", 198, "$ppcmd");
    }

    if ($dmpout[1] eq "on" && $dmpout[0] eq "dmp_monitor_osevent" && $PP == 1 )  {
        $ppcmd = "$VXDDLADM stop eventsource";
        VCSAG_LOG_MSG("I", "EMC PowerPath detected, dmp_monitor_osevent on, stopping eventsource:\n$ppcmd", 199, "$ppcmd");
    }


    if ($PP) {
        $cmd = "$POWERMT check_registration";
        my ($status, $output) = do_command($cmd);
        if ( (!$status) && (($output =~ /Capabilities: All/i ) && ($output =~ /Product: PowerPath/i)) ) {
            $PP = 1;
        } else {
            $PP = 0;
        }
    }
    return $PP;
}


#
# ScanDev - Scan devices after a snapshot has been taken to detect the newly visible snapshot luns
#

sub ScanDev {
    my ($self) = shift;
    my $vxflag = shift;
    my $cmd;
    my ($status, $output);

    if ($SYSTEM eq "Linux") {
        my @entries;
        my $entry;
        my ($kernel_release) = split('\n', `uname -r`);
        my $test;
        my ($linscan, $linlog);
        if ($kernel_release =~ /2.6/) {
            if ($vxflag == 0 ) {
                VCSAG_LOG_MSG("I", "ScanDev:0: Refreshing linux device details. Please be patient.", 154);
                @entries = split('\s+', `ls //sys//class//scsi_host`);
                foreach $entry (@entries)
                {
                    VCSAG_LOG_MSG("I", "Issuing FC scan command on host [$entry].", 155, "$entry");
                    $test = "echo '- - -' > /sys/class/scsi_host/$entry/scan";
                    system($test);
                    VCSAG_LOG_MSG("I", "Done. Waiting for 5 secs.", 156);
                    sleep(5);
                }
                if (-f "/usr/bin/lsscsi") {
                    $linscan = "/usr/bin/lsscsi | $GREP DGC";
                    $linlog = `$linscan`;
                    VCSAG_LOG_MSG("I", "\n\nLinux device tree output :$linscan\n$linlog\n", 200, "$linscan", "$linlog");
                }
            }
        }

        if ($vxflag == 2 ) {
            VCSAG_LOG_MSG("I", "ScanDev:2: Issuing linux device scan command. Please be patient.", 157);
            @entries = split('\s+', `ls //sys//class//scsi_host`);
            foreach $entry (@entries)
            {
                VCSAG_LOG_MSG("I", "Issuing FC scan command on host [$entry].", 158, "$entry");
                $test = "echo '- - -' > /sys/class/scsi_host/$entry/scan";
                system($test);
                VCSAG_LOG_MSG("I", "Done. Waiting for 5 secs.", 159);
                sleep(5);
            }
        }


        if ($vxflag == 1 ) {
            if (-f "/usr/bin/lsscsi") {
                $linscan = "/usr/bin/lsscsi | $GREP DGC";
                $linlog = `$linscan`;
                VCSAG_LOG_MSG("I", "\n\nLinux device tree output :$linscan\n$linlog\n", 163, "$linscan", "$linlog");
            }
        }

    }
    elsif ($SYSTEM eq "AIX")
    {
        $cmd="$CFGMGR >/dev/null 2>&1";
        VCSAG_LOG_MSG("I", "Issuing cmd: [$cmd]", 100, "$cmd");
        system($cmd);
    }
    elsif ($SYSTEM eq "HP-UX")
    {
        $cmd="$IOSCAN";
        VCSAG_LOG_MSG("N", "ScanDev issuing HP-UX command: $cmd", 101, "$cmd");
        ($status, $output) = do_command($cmd);
        $cmd = "$VXDCTL enable";
    }
    elsif ($SYSTEM eq "SunOS")
    {

        my $PP = is_emc_powerpath_enable();


        if ($vxflag == 2 ) {

            push(@CMDS, "$CFGADM -alo show_FCP_dev");
            push(@CMDS, "$DEVFSADM -Cvc disk");
            my $ret = run_cmds(@CMDS);
            return 0;
        }

        my ($ppcmd, $ppcmdout_for_log, $pplog);;
        if ($PP) {
            VCSAG_LOG_MSG("I", "EMC PowerPath configured.\n\n",133);

            VCSAG_LOG_MSG("I", "Validate EMC PowerPath paths.\n",133);
            $ppcmd = "$POWERMT display";
            $ppcmdout_for_log = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP LUN Count: $ppcmd\n$ppcmdout_for_log\n", 138, "$ppcmd", "$ppcmdout_for_log");

            $ppcmd = "$POWERCF -q";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP command ouput :\n$ppcmd\n$pplog\n", 201, "$ppcmd", "$pplog");

            $ppcmd = "yes | $POWERMT restore dev=all";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP command: $ppcmd\n$pplog\n", 202, "$ppcmd", "$pplog");

            $ppcmd = "$POWERMT check dev=all class=clariion force";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP command: $ppcmd\n$pplog\n", 203, "$ppcmd", "$pplog");

            $ppcmd = "$POWERMT display";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP LUN Count: $ppcmd\n$pplog\n", 204, "$ppcmd", "$pplog");

            $ppcmd = "$POWERCF -q";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP command ouput :\n$ppcmd\n$pplog\n", 205, "$ppcmd", "$pplog");
        }


        if ( ($vxflag == 3 ) && $PP) {
            $PP = $self->cleanup_cfg_stack();
            if ($PP) {
                return $PP;
            }

            $PP = $self->restore_emc_stack();
            if ($PP) {
                return $PP;
            }

            return 0;
        }

        if ( ($vxflag == 1 ) && $PP) {
            $ppcmd = "$POWERCF -q";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP restore command ouput :\n$ppcmd\n$pplog\n", 206, "$ppcmd", "$pplog");

            $ppcmd = "yes | $POWERMT restore dev=all";
            $ppcmdout_for_log = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP restore command: $ppcmd\n$ppcmdout_for_log\n", 207, "$ppcmd", "$ppcmdout_for_log");

            $ppcmd = "$POWERMT check dev=all class=clariion force";
            $ppcmdout_for_log = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP restore command: $ppcmd\n$ppcmdout_for_log\n", 208, "$ppcmd", "$ppcmdout_for_log");

            $ppcmd = "$POWERCF -q";
            $pplog = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP restore command ouput :\n$ppcmd\n$pplog\n", 209, "$ppcmd", "$pplog");

            $ppcmd = "$POWERMT display";
            $ppcmdout_for_log = `$ppcmd`;
            VCSAG_LOG_MSG("I", "PP LUN Count: $ppcmd\n$ppcmdout_for_log\n", 210, "$ppcmd", "$ppcmdout_for_log");
        }
    } elsif ($SYSTEM eq "Windows NT") {
        $cmd = "$VXDISK rescan >nul 2>&1";
        VCSAG_LOG_MSG("I", "ScanDev: Issuing cmd: [$cmd]", 100, "$cmd");
        system($cmd);
        return 0;
    }
    $cmd = "/usr/sbin/vxdctl enable";
    VCSAG_LOG_MSG("I", "ScanDev: Issuing cmd: [$cmd]", 100, "$cmd");
    system($cmd);

    $cmd = "/usr/sbin/vxdisk scandisks";
    VCSAG_LOG_MSG("I", "ScanDev: Issuing cmd: [$cmd]", 100, "$cmd");
    system($cmd);

    $cmd = "$VXDISK scandisks";
    VCSAG_LOG_MSG("I", "ScanDev: issuing VM command: $cmd\n\n", 211, "$cmd");
    ($status, $output) = do_command($cmd);
    if ( $output =~ /Data Corruption Protection Activated/) {
        VCSAG_LOG_MSG("E", "Data Corruption Protection Activated - User Corrective Action Needed. Hence Exiting after cleaning snapshot devices.", 137);
        $self->resynch(1);
        exit 1;
    }

    $cmd = "/usr/sbin/vxddladm assign names";
    VCSAG_LOG_MSG("I", "ScanDev: Issuing cmd: [$cmd]", 100, "$cmd");
    system($cmd);

    $cmd = "$VXDISK -o alldgs list";
    VCSAG_LOG_MSG("I", "ScanDev: Issuing cmd: [$cmd]", 100, "$cmd");
    my $op = `$cmd`;
    VCSAG_LOG_MSG("I", "ScanDev: output: [$op]", 100, "$op");

    return 0;
}

sub GetIPVersion {
    my $host = shift;
    my $ip_address;

    if ( ! defined &Net::IP::ip_is_ipv4 ) {
        return 4
    }

    if($host =~/^[0-9]/) {
        $ip_address = $host;
    } else {
        $ip_address = Socket::inet_ntoa (Socket::inet_aton($host));
    }

    if (Net::IP::ip_is_ipv4($ip_address)) {
        return 4;
    }
    elsif (Net::IP::ip_is_ipv6($ip_address)) {
        return 6;
    }
    else {
        return -1;
    }
}


#
# PingTest - ping host to find out if it is alive
# returns
#  0 if alive
#  1 otherwise
#
sub PingTest {
    my ($host) = @_;
    my $cmd;
    my $ret = 1;

    my $ipversion = GetIPVersion($host);

    if ( $ipversion == 4 ) {
        $ret = Ping4Test($host);
    } elsif ( $ipversion == 6 ) {
        $ret = Ping6Test($host);
    }
    return $ret;
}


sub Ping4Test {
    my ($host) = @_;
    my $cmd;

    if ($SYSTEM eq "AIX") {
        $cmd="ping -c 1 $host";
    } elsif ($SYSTEM eq "HP-UX") {
        $cmd="ping $host -n 1";
    } elsif ($SYSTEM eq "Linux") {
        $cmd="ping -c 1 $host";
    } elsif ($SYSTEM eq "SunOS") {
        $cmd="ping $host 1";
    } elsif ($SYSTEM eq "Windows NT") {
        $cmd="ping $host";
    }
    my ($status, $output) = do_command($cmd);
    return 1 if ($output =~ /no answer from/);
    return 1 if ($output =~ /unknown host/);
    return 1 if ($output =~ /could not find host/);
    return 1 if ($output =~ /Request timed out/);
    return 1 if ($status);
    return 0;
}

sub Ping6Test {
    my ($host) = @_;
    my $cmd;

    if ($SYSTEM eq "AIX") {
        $cmd="ping -c 1 -a inet6 $host";
    } elsif ($SYSTEM eq "HP-UX") {
        $cmd="ping -f inet6 $host -n 1";
    } elsif ($SYSTEM eq "Linux") {
        $cmd="ping6 -c 1 $host";
    } elsif ($SYSTEM eq "SunOS") {
        $cmd="ping -A inet6 $host 1";
    } elsif ($SYSTEM eq "Windows NT") {
        $cmd="ping $host";
    }
    my ($status, $output) = do_command($cmd);
    return 1 if ($output =~ /no answer from/);
    return 1 if ($output =~ /unknown host/);
    return 1 if ($output =~ /could not find host/);
    return 1 if ($output =~ /Request timed out/);
    return 1 if ($status);
    return 0;
}

sub reenable_vxconfigd_on_slaves {
	my $self = shift;
	my $is_master = 0;
	my (@v1) = ();
	my $nodename = "";
	my $ret = 0;
	my $reenable_failed = 0;

	if ($self->{'rac'} == 0) {
		return 0;
	}

	# RAC cluster - slave does nothing here
	$is_master = ArrayRep::is_cvm_master();
	if ($is_master == -1) {
		return 1;
	}
	elsif ($is_master == 0) {
		# Slave node: dont do anything
		return 0;
    }

	# get the list of all nodes in the cvm cluster
	my (@vxclus) = split('\n', `$VXCLUSTADM nidmap 2>$DEVNULL`);
	foreach my $v (@vxclus) {
		if (($v !~ /Master/) && ($v !~ /^Name/)) {
			# for all slave nodes
			(@v1) = split('\s+', $v);
			$nodename = $v1[0];
			$ret = $self->reenable_vxconfigd($nodename);
			my $cmd = "$ag_i18n_inc::HARES -action $self->{'resname'} devscan -sys $nodename";
			VCSAG_LOG_MSG("I", "reenable_vxconfigd(): Issuing cmd: [$cmd]", 100, "$cmd");
			`$cmd`;
			$reenable_failed = $?;
		}
	}
	if ($reenable_failed == 1) {
		VCSAG_LOG_MSG("W",
	"The action to reenable the volume configuration daemon failed on atleast one slave node in the CVM cluster. This may cause the CVM diskgroup to not be imported successfully.", 140);
	}
}

sub reenable_vxconfigd {
	my ($self, $nodename) = @_;
	my $reenable_failed = 0;

	return $reenable_failed;
}


#
# Arguments to the subroutine:
# ----------------------------
# dgname: The string comprising of the original diskgroups names
# (without the _fd appended to the names) separated by a " "
# case: Type of Firedrill: gold/silver/bronze
# tagname: The value of the tag set on the disks on which the diskgroup
#   is to be imported
# is_rac_cluster: 1 if this is a RAC environment in which the agent is
#   running; 0 otherwise
#
# Return Values:
# --------------
# 0 on success;
# 1 otherwise and in this case, an appropriate error message is logged
#
# Functionality:
# --------------
# This subroutine checks that after the disks have been setup right,
# the diskgroups can be successfully imported on the system. If the
# subroutine encounters error in attempting to import/deport the
# diskgroups, it returns a failure code (1) along with logging the
# appropriate error message.
#

sub vxvm_test_diskgroup_import {
    my ($self, $case, $tagname, $is_rac_cluster) = @_;
    my $newdg;
    my $options = "";
    my $rac_options = "";

    #
    # $options contains the options to be used during the diskgroup
    # import in a non-RAC env whereas $rac_options contains the
    # options to be used during the diskgroup import in a RAC env
    #

    my (@cmds) = ();
    my $ret = 0;

    if ($case eq "gold") {
        #
        # GOLD firedrill
        #
        $options = "-C -o useclonedev=on -o updateid -o tag=$tagname";
        $rac_options = "-s -C -o updateid -o tag=$tagname";
    } else {
        #
        # BRONZE or SILVER Firedrill
        #
        $options = "-C -o updateid";
        $rac_options = "-s -C -o updateid";
    }

    foreach my $dgname (@{$self->{'dgname'}}) {
        $newdg = "$dgname" . "_fd";
        #
        # Import the dg as a non-shared dg first. This is a
        # workaround for a bug in CVM. Refer to etrack incident
        # 1053529 for more info on this bug.
        #
        push(@cmds, "vxdg $options -n $newdg import -g $dgname");
        my $cmd = "vxdg $options -n $newdg import -g $dgname";

        my $op = `$cmd`;
        #$ret = run_cmds(\@cmds);
    if ($?) {
        VCSAG_LOG_MSG("E", "Cannot import disk group $dgname in pitc $case; exiting",
            60003, "$dgname", "$case");
        return 1;
    }
}

return 0;
}

1;
